/*
 * Ben NanoNote and Arduino USB Host shield communication.
 *
 * Copyright 2013 Paul Boddie
 *
 * SPI functions derived from those in lib/atben.c by Werner Almesberger:
 *
 * Copyright 2010-2011 Werner Almesberger
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <ubb/ubb.h>
#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <usb.h>
#include <unistd.h>
#include <string.h>
#include <endian.h>

/* Found in Python's asdl.h. */

#ifndef __cplusplus
typedef enum {false, true} bool;
#endif

/* Initialisation states. */

typedef enum
{
    MAX_DEVSTATE_INIT = 0,
    MAX_DEVSTATE_CONNECTED = 1,
    MAX_DEVSTATE_RESET = 2,
    MAX_DEVSTATE_INSPECTED = 3,
    MAX_DEVSTATE_RESET_AGAIN = 4,
    MAX_DEVSTATE_READY = 5
} max_devstate;

/* Device details. */

typedef struct
{
    bool in_toggle, out_toggle;
    struct usb_device_descriptor desc;
} max_device;

/* Pin assignments:
 *
 * Sniffer  UBB   Shield
 * -------  ----  ------
 * DAT2     DAT2       9 (INT)
 * CD       DAT3      10 (SS)
 * CMD      CMD        7 (RESET)
 * VCC      VDD      VIN
 * CLK      CLK       13 (SCLK)
 * GND      GND      GND
 * DAT0     DAT0      11 (MOSI)
 * DAT1     DAT1      12 (MISO)
 *                     8 (GPX) (not assigned)
 */

#define MAX_RESET UBB_CMD
#define MAX_SCLK  UBB_CLK
#define MAX_MOSI  UBB_DAT0
#define MAX_MISO  UBB_DAT1
#define MAX_INT   UBB_DAT2
#define MAX_SS    UBB_DAT3

/* MAX3421E definitions. */

#define MAX_REG_READ                0x00
#define MAX_REG_WRITE               0x02

#define MAX_REG_RCVFIFO             1
#define MAX_REG_SNDFIFO             2
#define MAX_REG_SUDFIFO             4
#define MAX_REG_RCVBC               6
#define MAX_REG_SNDBC               7
#define MAX_REG_USBIRQ              13
#define MAX_REG_USBIEN              14
#define MAX_REG_USBCTL              15
#define MAX_REG_CPUCTL              16
#define MAX_REG_PINCTL              17
#define MAX_REG_REVISION            18
#define MAX_REG_HIRQ                25
#define MAX_REG_HIEN                26
#define MAX_REG_MODE                27
#define MAX_REG_PERADDR             28
#define MAX_REG_HCTL                29
#define MAX_REG_HXFR                30
#define MAX_REG_HRSL                31

#define MAX_USBIRQ_OSCOKIRQ         1
#define MAX_USBIRQ_NOVBUSIRQ        32
#define MAX_USBIRQ_VBUSIRQ          64

#define MAX_USBCTL_PWRDOWN          16
#define MAX_USBCTL_CHIPRES          32

#define MAX_CPUCTL_IE               1

#define MAX_PINCTL_POSINT_LOW       0
#define MAX_PINCTL_POSINT_HIGH      4
#define MAX_PINCTL_INTLEVEL_EDGE    0
#define MAX_PINCTL_INTLEVEL_LEVEL   8
#define MAX_PINCTL_FDUPSPI_HALF     0
#define MAX_PINCTL_FDUPSPI_FULL     16

#define MAX_HIRQ_BUSEVENTIRQ        1
#define MAX_HIRQ_RWUIRQ             2
#define MAX_HIRQ_RCVDAVIRQ          4
#define MAX_HIRQ_SNDBAVIRQ          8
#define MAX_HIRQ_SUSDNIRQ           16
#define MAX_HIRQ_CONDETIRQ          32
#define MAX_HIRQ_FRAMEIRQ           64
#define MAX_HIRQ_HXFRDNIRQ          128

#define MAX_HIEN_BUSEVENTIE         1
#define MAX_HIEN_CONDETIE           32
#define MAX_HIEN_FRAMEIE            64

#define MAX_MODE_PERIPHERAL         0
#define MAX_MODE_HOST               1
#define MAX_MODE_LOWSPEED           2
#define MAX_MODE_SOFKAENAB          8
#define MAX_MODE_SEPIRQ_OFF         0
#define MAX_MODE_SEPIRQ_ON          16
#define MAX_MODE_DMPULLDN           64
#define MAX_MODE_DPPULLDN           128

#define MAX_MODE_HOST_ENABLED       MAX_MODE_HOST | MAX_MODE_SEPIRQ_OFF | MAX_MODE_DMPULLDN | MAX_MODE_DPPULLDN
#define MAX_MODE_HOST_FULLSPEED     MAX_MODE_HOST_ENABLED
#define MAX_MODE_HOST_LOWSPEED      MAX_MODE_HOST_ENABLED | MAX_MODE_LOWSPEED

#define MAX_HCTL_BUSRST             1
#define MAX_HCTL_SAMPLEBUS          4
#define MAX_HCTL_RCVTOG0            16
#define MAX_HCTL_RCVTOG1            32
#define MAX_HCTL_SNDTOG0            64
#define MAX_HCTL_SNDTOG1            128

#define MAX_HXFR_SETUP              16
#define MAX_HXFR_OUTNIN             32
#define MAX_HXFR_HS                 128

#define MAX_HRSL_JSTATUS            128
#define MAX_HRSL_KSTATUS            64
#define MAX_HRSL_SNDTOGRD           32
#define MAX_HRSL_RCVTOGRD           16
#define MAX_HRSL_HRSLT              15

#define max_reg(n) ((uint8_t) (n << 3))
#define max_reg_read(n) (max_reg(n) | MAX_REG_READ)
#define max_reg_write(n) (max_reg(n) | MAX_REG_WRITE)

void spi_begin()
{
    CLR(MAX_SS);
}

void spi_end()
{
    SET(MAX_SS);
}

/**
 * Send the given value via MOSI while receiving a value via MISO.
 * This requires full-duplex SPI and will produce a status value for the first
 * value sent (the command).
 */
uint8_t spi_sendrecv(uint8_t v)
{
    uint8_t result = 0;
    uint8_t mask;

    for (mask = 0x80; mask; mask >>= 1)
    {
        if (v & mask)
        {
            #ifdef DEBUG
            printf("1");
            #endif
            SET(MAX_MOSI);
        }
        else
        {
            #ifdef DEBUG
            printf("0");
            #endif
            CLR(MAX_MOSI);
        }

        /* Wait for stable output signal. */

        SET(MAX_SCLK);

        if (PIN(MAX_MISO))
            result |= mask;

        CLR(MAX_SCLK);
    }

    #ifdef DEBUG
    printf("\n");
    #endif
    return result;
}

uint8_t max_read(uint8_t reg, uint8_t *status)
{
    uint8_t result = 0, tmpstatus = 0;

    tmpstatus = 0;

    spi_begin();
    tmpstatus = spi_sendrecv(max_reg_read(reg));
    result = spi_sendrecv(0);
    spi_end();

    if (status != NULL)
        *status = tmpstatus;

    return result;
}

uint8_t max_write(uint8_t reg, uint8_t value)
{
    uint8_t status = 0;

    spi_begin();
    status = spi_sendrecv(max_reg_write(reg));
    spi_sendrecv(value);
    spi_end();

    return status;
}

/**
 * Return whether data can be sent.
 */
bool max_can_send(uint8_t *status)
{
    if (status == NULL)
        return max_read(MAX_REG_HIRQ, NULL) & MAX_HIRQ_SNDBAVIRQ;
    else
        return *status & MAX_HIRQ_SNDBAVIRQ;
}

/**
 * Set the sending data toggle.
 */
void max_set_send_toggle(bool toggle)
{
    max_write(MAX_REG_HCTL, toggle ? MAX_HCTL_SNDTOG1 : MAX_HCTL_SNDTOG0);
}

/**
 * Return the sending data toggle.
 */
bool max_get_send_toggle()
{
    return (max_read(MAX_REG_HRSL, NULL) & MAX_HRSL_SNDTOGRD) != 0;
}

/**
 * Set the receiving data toggle.
 */
void max_set_recv_toggle(bool toggle)
{
    max_write(MAX_REG_HCTL, toggle ? MAX_HCTL_RCVTOG1 : MAX_HCTL_RCVTOG0);
}

/**
 * Return the receiving data toggle.
 */
bool max_get_recv_toggle()
{
    return (max_read(MAX_REG_HRSL, NULL) & MAX_HRSL_RCVTOGRD) != 0;
}

/**
 * Wait for handshake/timeout after a transfer.
 */
uint8_t max_wait_transfer(uint8_t status)
{
    while (!(status & MAX_HIRQ_HXFRDNIRQ))
    {
        status = max_read(MAX_REG_HIRQ, NULL);
    }

    max_write(MAX_REG_HIRQ, MAX_HIRQ_HXFRDNIRQ);
    return max_read(MAX_REG_HIRQ, NULL);
}

/**
 * Write the given data to the FIFO.
 */
void max_write_fifo(uint8_t endpoint, uint8_t *data, uint8_t len)
{
    uint8_t count;

    for (count = 0; count < len; count++)
    {
        max_write(endpoint ? MAX_REG_SNDFIFO : MAX_REG_SUDFIFO, data[count]);
    }

    if (endpoint)
        max_write(MAX_REG_SNDBC, len);
}

/**
 * Read the data from the FIFO.
 */
void max_read_fifo(uint8_t **data, uint8_t *len, uint8_t *datalimit)
{
    uint8_t count, received = max_read(MAX_REG_RCVBC, NULL);
    printf("Received %d bytes.\n", received);

    *len += received;

    for (count = 0; (count < received) && (*data < datalimit); count++)
    {
        *((*data)++) = max_read(MAX_REG_RCVFIFO, NULL);
    }
}

/**
 * Send a control request consisting of the given setup data.
 */
uint8_t max_control(uint8_t *setup)
{
    uint8_t status, hrsl;

    max_write_fifo(0, setup, 8);

    /* Initiate the transfer. */

    do
    {
        status = max_write(MAX_REG_HXFR, MAX_HXFR_SETUP);
        status = max_wait_transfer(status);
        hrsl = max_read(MAX_REG_HRSL, &status);
    }
    while (hrsl & MAX_HRSL_HRSLT);

    return status;
}

/**
 * Perform a status transaction as part of a larger control transaction.
 * The out parameter is used to indicate the kind of status transfer to be
 * performed and should be the inverse of the control transfer direction.
 */
uint8_t max_control_status(bool out)
{
    uint8_t status, hrsl;

    do
    {
        status = max_write(MAX_REG_HXFR, MAX_HXFR_HS | (out ? MAX_HXFR_OUTNIN : 0));
        status = max_wait_transfer(status);
        hrsl = max_read(MAX_REG_HRSL, &status);
    }
    while (hrsl & MAX_HRSL_HRSLT);

    return status;
}

/**
 * Send a request to the given endpoint, using the supplied data payload with
 * the given length, indicating the preserved toggle state of the endpoint
 * (which will be updated).
 */
uint8_t max_send(uint8_t endpoint, uint8_t *data, uint8_t len, bool *toggle)
{
    uint8_t status, hrsl = 0;

    max_write_fifo(endpoint, data, len);

    max_set_send_toggle(*toggle);

    /* Initiate the transfer. */

    do
    {
        status = max_write(MAX_REG_HXFR, endpoint | MAX_HXFR_OUTNIN);
        status = max_wait_transfer(status);

        /* Test for usable data. */

        if (status & MAX_HIRQ_SNDBAVIRQ)
        {
            hrsl = max_read(MAX_REG_HRSL, &status);

            if (!(hrsl & MAX_HRSL_HRSLT))
                break;
        }
    }
    while (true);

    *toggle = max_get_send_toggle();

    return status;
}

/**
 * Make a request for data from the given endpoint, collecting it in the
 * supplied buffer with the given length, indicating the preserved toggle state
 * of the endpoint (which will be updated). The length will be updated to
 * indicate the total length of the received data.
 */
bool max_recv(uint8_t endpoint, uint8_t *data, uint8_t *len, bool *toggle)
{
    uint8_t *datalimit = data + *len;
    uint8_t status, hrsl = 0;
    uint16_t attempt = 1024;

    max_set_recv_toggle(*toggle);

    /* Initiate the transfer. */

    do
    {
        status = max_write(MAX_REG_HXFR, endpoint);
        status = max_wait_transfer(status);

        /* Test for usable data. */

        hrsl = max_read(MAX_REG_HRSL, &status);

        attempt--;
    }
    while ((hrsl & MAX_HRSL_HRSLT) && attempt);

    if (!attempt)
    {
        printf("HRSL is %x\n", hrsl);
        return false;
    }

    while (status & MAX_HIRQ_RCVDAVIRQ)
    {
        max_read_fifo(&data, len, datalimit);

        /* Indicate that all data has been read. */

        max_write(MAX_REG_HIRQ, MAX_HIRQ_RCVDAVIRQ);
        status = max_read(MAX_REG_HIRQ, NULL);
    }

    *toggle = max_get_recv_toggle();

    return true;
}

void chipreset()
{
    printf("Resetting...\n");
    max_write(MAX_REG_USBCTL, MAX_USBCTL_CHIPRES);

    printf("Clearing the reset...\n");
    max_write(MAX_REG_USBCTL, 0);
}

uint8_t check()
{
    uint8_t oscillator;

    oscillator = max_read(MAX_REG_USBIRQ, NULL);

    return (oscillator & ~(MAX_USBIRQ_NOVBUSIRQ | MAX_USBIRQ_VBUSIRQ)) == MAX_USBIRQ_OSCOKIRQ;
}

uint8_t wait()
{
    uint16_t timeout = 1024;

    /* Wait for the oscillator before performing USB activity. */

    printf("Waiting...\n");

    while ((timeout > 0) && (!check()))
    {
        timeout--;
    }

    printf("Iterations remaining: %d\n", timeout);

    return timeout;
}

/**
 * Return whether the bus is ready to be sampled. The application note claims
 * that the SAMPLEBUS bit is cleared, but this does not seem to be the case.
 */
uint8_t samplebusready()
{
    uint8_t result;

    result = max_read(MAX_REG_HCTL, NULL);
    return result & MAX_HCTL_SAMPLEBUS;
}

void samplebus()
{
    max_write(MAX_REG_HCTL, MAX_HCTL_SAMPLEBUS);
    while (!samplebusready());
}

/**
 * Handle the connection or disconnection of a device, returning true if the
 * device is now connected or false otherwise. If the device is connected, a
 * bus reset is performed.
 */
bool devicechanged()
{
    uint8_t hrsl, mode;

    hrsl = max_read(MAX_REG_HRSL, NULL);
    mode = max_read(MAX_REG_MODE, NULL);

    if ((hrsl & MAX_HRSL_JSTATUS) && (hrsl & MAX_HRSL_KSTATUS))
    {
        printf("Bad device status.\n");
    }
    else if (!(hrsl & MAX_HRSL_JSTATUS) && !(hrsl & MAX_HRSL_KSTATUS))
    {
        printf("Device disconnected.\n");
    }
    else
    {
        printf("Device connected.\n");

        /* Low speed device when J and lowspeed have the same level.
           Since J and K should have opposing levels, K can be tested when
           lowspeed is low. */

        if (((hrsl & MAX_HRSL_JSTATUS) && (mode & MAX_MODE_LOWSPEED)) ||
            ((hrsl & MAX_HRSL_KSTATUS) && !(mode & MAX_MODE_LOWSPEED)))
        {
            printf("Device is low speed.\n");
            max_write(MAX_REG_MODE, MAX_MODE_HOST_LOWSPEED);
        }
        else
        {
            printf("Device is full speed.\n");
            max_write(MAX_REG_MODE, MAX_MODE_HOST_FULLSPEED);
        }

        /* Reset the device. */

        max_write(MAX_REG_HCTL, MAX_HCTL_BUSRST);
        return true;
    }

    return false;
}

void setup_packet(uint8_t *setup, uint8_t request_type, uint8_t request, uint16_t value, uint16_t index, uint16_t length)
{
    setup[0] = request_type;
    setup[1] = request;
    setup[2] = value & 0xff;
    setup[3] = value >> 8;
    setup[4] = index & 0xff;
    setup[5] = index >> 8;
    setup[6] = length & 0xff;
    setup[7] = length >> 8;
}

bool max_init_device(max_device *device)
{
    uint8_t data[64], len = 64, setup[8];

    printf("Sending control request to address 0, endpoint 0...\n");

    max_write(MAX_REG_PERADDR, 0);
    setup_packet(setup, USB_ENDPOINT_IN, USB_REQ_GET_DESCRIPTOR, USB_DT_DEVICE << 8, 0, USB_DT_DEVICE_SIZE);
    max_control(setup);
    device->in_toggle = true;
    if (!max_recv(0, data, &len, &device->in_toggle))
        return false;
    max_control_status(true);

    if (len >= sizeof(struct usb_device_descriptor))
    {
        memcpy(&device->desc, data, sizeof(struct usb_device_descriptor));
        device->desc.bcdUSB = le16toh(device->desc.bcdUSB);
        device->desc.idVendor = le16toh(device->desc.idVendor);
        device->desc.idProduct = le16toh(device->desc.idProduct);
        device->desc.bcdDevice = le16toh(device->desc.bcdDevice);

        printf("bLength: %d\n", device->desc.bLength);
        printf("bDescriptorType: %d\n", device->desc.bDescriptorType);
        printf("bcdUSB: %04x\n", device->desc.bcdUSB);
        printf("bDeviceClass: %d\n", device->desc.bDeviceClass);
        printf("bDeviceSubClass: %d\n", device->desc.bDeviceSubClass);
        printf("bDeviceProtocol: %d\n", device->desc.bDeviceProtocol);
        printf("bMaxPacketSize0: %d\n", device->desc.bMaxPacketSize0);
        printf("idVendor: %04x\n", device->desc.idVendor);
        printf("idProduct: %04x\n", device->desc.idProduct);
        printf("bcdDevice: %04x\n", device->desc.bcdDevice);
        printf("iManufacturer: %d\n", device->desc.iManufacturer);
        printf("iProduct: %d\n", device->desc.iProduct);
        printf("iSerialNumber: %d\n", device->desc.iSerialNumber);
        printf("bNumConfigurations: %d\n", device->desc.bNumConfigurations);

        /* Reset the device. */

        max_write(MAX_REG_HCTL, MAX_HCTL_BUSRST);
        return true;
    }

    return false;
}

void shutdown(int signum)
{
    printf("Closing...\n");
    ubb_close(0);
    exit(1);
}

int main(int argc, char *argv[])
{
    uint8_t status = 0, revision = 0;
    uint16_t count;
    bool bus_event, data_event, suspended_event, connection_event, frame_event;
    max_device device;
    max_devstate devstate = MAX_DEVSTATE_INIT;

    device.in_toggle = false;
    device.out_toggle = false;

    signal(SIGINT, &shutdown);

    if (ubb_open(0) < 0) {
        perror("ubb_open");
        return 1;
    }

    ubb_power(1);
    printf("Power on.\n");

    OUT(MAX_SS);
    OUT(MAX_MOSI);
    OUT(MAX_SCLK);
    OUT(MAX_RESET);
    IN(MAX_INT);
    IN(MAX_MISO);

    /* Initialise SPI. */
    /* Set SS# to 1. */

    SET(MAX_SS);
    CLR(MAX_MOSI);
    CLR(MAX_SCLK);
    SET(MAX_RESET);

    /* Initialise the MAX3421E. */

    /* Set full-duplex, interrupt signalling. */

    printf("Setting pin control...\n");
    max_write(MAX_REG_PINCTL, MAX_PINCTL_INTLEVEL_LEVEL | MAX_PINCTL_FDUPSPI_FULL);

    chipreset();
    printf("Ready? %d\n", wait());

    /* Check various registers. */

    printf("Mode: %x\n", max_read(MAX_REG_MODE, &status));
    printf("IRQ: %x\n", max_read(MAX_REG_HIRQ, &status));

    /* Set host mode. */

    printf("Setting mode...\n");
    status = max_write(MAX_REG_MODE, MAX_MODE_HOST_ENABLED);

    printf("Setting INT signalling...\n");
    status = max_write(MAX_REG_CPUCTL, MAX_CPUCTL_IE);

    printf("Setting event signalling...\n");
    status = max_write(MAX_REG_HIEN, MAX_HIEN_CONDETIE | MAX_HIEN_FRAMEIE | MAX_HIEN_BUSEVENTIE);

    /* Check various registers. */

    printf("Mode: %x\n", max_read(MAX_REG_MODE, &status));
    printf("IRQ: %x\n", max_read(MAX_REG_HIRQ, &status));
    printf("IE: %x\n", max_read(MAX_REG_HIEN, &status));
    printf("CPU: %x\n", max_read(MAX_REG_CPUCTL, &status));
    printf("Pin: %x\n", max_read(MAX_REG_PINCTL, &status));
    printf("USBIRQ: %x\n", max_read(MAX_REG_USBIRQ, &status));
    printf("USBIE: %x\n", max_read(MAX_REG_USBIEN, &status));

    /* Read from the REVISION register. */

    printf("Reading...\n");
    revision = max_read(MAX_REG_REVISION, &status);
    printf("Revision = %x\n", revision);

    /* Sample the bus for any devices that are already connected. */

    samplebus();

    if (devicechanged())
    {
        devstate = MAX_DEVSTATE_CONNECTED;
    }

    /* Enter the main processing loop. */

    for (count = 0; count <= 65535; count++)
    {
        if (!PIN(MAX_INT))
        {
            /* Obtain interrupt conditions. */

            status = max_read(MAX_REG_HIRQ, NULL);

            bus_event = status & MAX_HIRQ_BUSEVENTIRQ;
            data_event = status & MAX_HIRQ_RCVDAVIRQ;
            suspended_event = status & MAX_HIRQ_SUSDNIRQ;
            connection_event = status & MAX_HIRQ_CONDETIRQ;
            frame_event = status & MAX_HIRQ_FRAMEIRQ;

            #ifdef DEBUG
            if (bus_event) printf("Bus ");
            if (data_event) printf("Data ");
            if (suspended_event) printf("Suspended ");
            if (connection_event) printf("Connection ");
            if (frame_event) printf("Frame ");
            printf("\n");
            #endif

            /* Acknowledge the interrupts. */

            max_write(MAX_REG_HIRQ, status);

            /* Detect device connection/disconnection. */

            if ((devstate == MAX_DEVSTATE_INIT) && connection_event && devicechanged())
            {
                devstate = MAX_DEVSTATE_CONNECTED;
                printf("CONNECTED\n");
            }

            /* Handle device reset initiation. */

            else if ((devstate == MAX_DEVSTATE_CONNECTED) && bus_event)
            {
                max_write(MAX_REG_MODE, max_read(MAX_REG_MODE, NULL) | MAX_MODE_SOFKAENAB);
                devstate = MAX_DEVSTATE_RESET;
                printf("RESET\n");
            }

            /* Handle device reset completion, getting device details and
               initiating a second reset. */

            else if ((devstate == MAX_DEVSTATE_RESET) && frame_event && max_can_send(&status))
            {
                if (max_init_device(&device))
                {
                    devstate = MAX_DEVSTATE_INSPECTED;
                    printf("INSPECTED\n");
                }
                else
                    printf("FAILED: RESET -> INSPECTED\n");
            }

            /* Handle the second device reset initiation. */

            else if ((devstate == MAX_DEVSTATE_INSPECTED) && bus_event)
            {
                devstate = MAX_DEVSTATE_RESET_AGAIN;
                printf("RESET AGAIN\n");
            }

            /* Handle the second device reset completion, initiating communications. */

            else if ((devstate == MAX_DEVSTATE_RESET_AGAIN) && frame_event && max_can_send(&status))
            {
                max_write(MAX_REG_MODE, max_read(MAX_REG_MODE, NULL) & ~MAX_MODE_SOFKAENAB);
                devstate = MAX_DEVSTATE_READY;
                printf("READY\n");
            }

            /* Handle device disconnection. */

            else if ((devstate != MAX_DEVSTATE_INIT) && connection_event)
            {
                devstate = MAX_DEVSTATE_INIT;
                printf("INIT\n");
            }
        }
    }

    printf("Closing...\n");
    ubb_close(0);

    return 0;
}
